/***********************************************************************************************************************
 *
 * blueBill Core - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://kenai.com/hg/bluebill~core-src
 *
 **********************************************************************************************************************/
package it.tidalwave.observation;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.util.Date;
import java.io.Serializable;
import it.tidalwave.util.As;
import it.tidalwave.util.Initializer;

/***********************************************************************************************************************
 *
 * Represents an observation, made at a certain time and {@link Location}, by certain {@link Observer}s and composed of 
 * one or more {@link ObservationItem}s.
 * 
 * @stereotype Datum
 * @composed "1" <find> "*" ObservationItem
 * @composed "*" <find> "0..1" Observer
 * @composed "*" - "0..1" Location
 * @composed "*" - "0..1" Source
 * @navassoc - "<create>" - Observation.ChangeBuilder
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public interface Observation extends As, Serializable
  {
    public static final Class<Observation> Observation = Observation.class;

    /*******************************************************************************************************************
     *
     * For internal purposes only. Do not use directly.
     *
     ******************************************************************************************************************/
    public static interface BaseBuilder<T>
      {
        /***************************************************************************************************************
         *
         * Sets the date of the {@link Observation} to create.
         *
         * @param  date  the date
         * @return       the {@code Builder}
         *
         **************************************************************************************************************/
        @Nonnull
        public T at (@Nonnull Date date);

        /***************************************************************************************************************
         *
         * Sets the {@code Location} of the {@link Observation} to create.
         *
         * @param  location   the {@code Location}
         * @return            the {@code Builder}
         *
         **************************************************************************************************************/
        @Nonnull
        public T at (@Nonnull Location location);

        /***************************************************************************************************************
         *
         * Adds an {@code Observable} to the {@link Observation} to create.
         *
         * @param  observable   the {@code Observable}
         * @param  cardinality  the {@code Cardinality}
         * @return              the {@code Builder}
         *
         **************************************************************************************************************/
        @Nonnull
        public T of (@Nonnull Observable observable,
                     @Nonnull Cardinality cardinality);

        /***************************************************************************************************************
         *
         * Adds an {@code Observable} to the {@link Observation} to create.
         *
         * @param  observable   the {@code Observable}
         * @param  cardinality  the {@code Cardinality}
         * @param  initializer  
         * @return              the {@code Builder}
         *
         **************************************************************************************************************/
        @Nonnull
        public T of (@Nonnull Observable observable,
                     @Nonnull Cardinality cardinality,
                     @Nonnull Initializer<ObservationItem> initializer);

        /***************************************************************************************************************
         *
         * Sets the {@code Observer} of the {@link Observation} to create.
         *
         * @param  observer   the {@code Observer}
         * @return            the {@code Builder}
         *
         **************************************************************************************************************/
        @Nonnull
        public T by (@Nonnull Observer observer);

        /***************************************************************************************************************
         *
         * Sets the {@code Source} of the {@link Observation} to create.
         *
         * @param  source     the {@code Source}
         * @return            the {@code Builder}
         *
         **************************************************************************************************************/
        @Nonnull
        public T from (@Nonnull Source source);

        /***************************************************************************************************************
         *
         * Sets an extra capability of the {@link Observation} to create.
         *
         * @param  extra      an extra capability
         * @return            the {@code Builder}
         *
         **************************************************************************************************************/
        @Nonnull
        public T with (@Nonnull Object extra);

        /***************************************************************************************************************
         *
         *
         **************************************************************************************************************/
        @Nonnegative
        public int getItemCount();
      }

    /*******************************************************************************************************************
     *
     * A builder used to create new {@link Observation}s. This class implements
     * the <a href="http://en.wikipedia.org/wiki/Fluent_interface">Fluent Interface</a>
     * pattern. A typical usage is:
     *
     * <pre>
     * ObservationSet observationSet = ...;
     *
     * Date when = ...;
     * Location camargue = observationSet.findOrCreate(Location, "Camargue", ...);
     * Observer johnDoe = observationSet.findOrCreate(Observer, "John Doe", ...);
     * Observable flamingo = observationSet.findOrCreate(Observable, "Flamingo", ...);
     *
     * Observation observation = observationSet.createObservation().
     *                                          at(when).
     *                                          at(camargue).
     *                                          with(flamingo, Cardinality.valueOf(4)).
     *                                          by(johnDoe).
     *                                          build();
     * </pre>
     *
     * @navassoc - <create> - Observation 
     * 
     ******************************************************************************************************************/
    public static interface Builder extends BaseBuilder<Builder>
      {
        /***************************************************************************************************************
         *
         * Creates a new {@link Observation}.
         *
         * @return  the new {@code Observation}
         *
         **************************************************************************************************************/
        @Nonnull
        public Observation build();
      }

    /*******************************************************************************************************************
     *
     * A builder used to modify existing {@link Observation}s. Since {@code Observation} is an immutable class, this
     * builder creates a new instance which replaces the original one.
     * 
     * @navassoc - <create> - Observation 
     *
     ******************************************************************************************************************/
    public static interface ChangeBuilder extends BaseBuilder<ChangeBuilder>
      {
        /***************************************************************************************************************
         *
         * Removes the given {@link ObservationItem} from 
         *
         **************************************************************************************************************/
        @Nonnull
        public ChangeBuilder without (@Nonnull ObservationItem observationItem);

        /***************************************************************************************************************
         *
         * Applies the changes and creates a new {@link Observation} which replaces the original one.
         *
         * @return  the new {@code Observation}
         *
         **************************************************************************************************************/
        @Nonnull
        public Observation apply(); 
      }

    /*******************************************************************************************************************
     *
     * Returns the time at which this {@code Observation} has been made.
     *
     * @return  the time
     *
     ******************************************************************************************************************/
    @CheckForNull
    public Date getDate();

    /*******************************************************************************************************************
     *
     * Returns the {@link Location} where this {@code Observation} has been made.
     *
     * @return  the {@code Location}
     *
     ******************************************************************************************************************/
    @Nonnull
    public Location getLocation();

    /*******************************************************************************************************************
     *
     * Returns the {@link Source} from which this {@code Observation} comes.
     *
     * @return  the {@code Source}
     *
     ******************************************************************************************************************/
    @Nonnull
    public Source getSource();

    /*******************************************************************************************************************
     *
     * Returns a {@link Finder} for {@link ObservationItem}s that are part of
     * this {@code Observation}.
     *
     * FIXME: use a single find(Class)
     * @return  the {@link Finder}
     *
     ******************************************************************************************************************/
    @Nonnull
    public Finder<ObservationItem> findObservationItems();

    /*******************************************************************************************************************
     *
     * Returns a {@link Finder} for {@link Observer}s that made
     * this {@code Observation}.
     * FIXME: use a single find(Class)
     *
     * @return  the {@link Finder}
     *
     ******************************************************************************************************************/
    @Nonnull
    public Finder<Observer> findObservers();

    /*******************************************************************************************************************
     *
     * Creates a Builder that allows to modify this observation and replace the original.
     *
     ******************************************************************************************************************/
    @Nonnull
    public ChangeBuilder change();
  }
