/***********************************************************************************************************************
 *
 * blueBill Core - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://kenai.com/hg/bluebill~core-src
 *
 **********************************************************************************************************************/
package it.tidalwave.observation;

import javax.annotation.Nonnull;
import org.openide.util.Lookup;
import it.tidalwave.util.As;
import it.tidalwave.util.Id;
import it.tidalwave.util.Initializer;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.observation.Observation.Builder;
import it.tidalwave.observation.event.ObservationListener;

/***********************************************************************************************************************
 *
 * Represents a set of {@link Observation}s. 
 *
 * @stereotype Datum
 * @composed "1" - "*" Observation
 * @navassoc - "<create>\n<find>" - Observer
 * @navassoc - "<create>\n<find>" - Observable
 * @navassoc - "<create>\n<find>" - Source
 * @navassoc - "<create>\n<find>" - Location
 * @navassoc - "<create>\n<find>" - TextNote
 * @composed "1" "call" "0..*" ObservationListener
 * @navassoc - "<create>" - Observation.Builder
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 * 
 **********************************************************************************************************************/
public interface ObservationSet extends Lookup.Provider, As
  {
    /*******************************************************************************************************************
     *
     * Finds or create an object in the current {@code ObservationSet}.
     *
     * @param  type         the type of the object to find or create
     * @param  displayName  the display name of the object to find or create
     * @return              the object
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> T findOrCreate (@Nonnull Class<T> type,
                               @Nonnull String displayName);

    /*******************************************************************************************************************
     *
     * Finds or create an object in the current {@code ObservationSet}.
     *
     * @param  clazz        the type of the object to find or create
     * @param  displayName  the display name of the object to find or create
     * @param  initializer  an initializer to set the initial state of the
     *                      object in case it is created
     * @return              the object
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> T findOrCreate (@Nonnull Class<T> type,
                               @Nonnull String displayName,
                               @Nonnull Initializer<T> initializer);

    /*******************************************************************************************************************
     *
     * Finds an object given its id. FIXME: drop, use find() instead.
     *
     * @param  type               the type of the object to find
     * @param  id                 the id of the object to find
     * @throws NotFoundException  if the id doesn't match to any object
     * @return                    the object
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> T findById (@Nonnull Class<T> type, @Nonnull Id id)
      throws NotFoundException;

    /*******************************************************************************************************************
     *
     * Returns a {@link Finder} for performing a query of the given type.
     *
     * @param  type               the type of the objects to find
     * @return                    the {@code Finder}
     *
     ******************************************************************************************************************/
    @Nonnull
    public <T> Finder<T> find (@Nonnull Class<T> type);

    /*******************************************************************************************************************
     *
     * Returns a {@link Builder} for creating an {@link Observation}.
     *
     * @return  the {@code Builder}
     *
     ******************************************************************************************************************/
    @Nonnull
    public Observation.Builder createObservation();

    /*******************************************************************************************************************
     *
     * Removes all the {@link Observation}s.
     *
     ******************************************************************************************************************/
    public void clear();

    /*******************************************************************************************************************
     *
     * Registers a listener for being notified about operations on {@link Observation}s.
     * 
     * @param  listener  the listener to register
     *
     ******************************************************************************************************************/
    public void addListener (@Nonnull ObservationListener listener);

    /*******************************************************************************************************************
     *
     * Unregisters a listener for being notified about operations on {@link Observation}s.
     * 
     * @param  listener  the listener to unregister
     *
     ******************************************************************************************************************/
    public void removeListener (@Nonnull ObservationListener listener);

    /*******************************************************************************************************************
     *
     * @deprecated 
     *
     ******************************************************************************************************************/
    @Nonnull
    @Deprecated // TODO replace with findOrCreate()
    public Source createSource (@Nonnull String uri,
                                @Nonnull String displayName);
  }
