/***********************************************************************************************************************
 *
 * blueBill Core - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it
 * SCM: https://kenai.com/hg/bluebill~core-src
 *
 **********************************************************************************************************************/
package it.tidalwave.observation.spi;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;
import org.slf4j.Logger;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.Serializable;
import it.tidalwave.observation.Observation;
import it.tidalwave.observation.ObservationSet;
import it.tidalwave.observation.event.ObservationEvent;
import it.tidalwave.observation.event.ObservationListener;
import org.slf4j.LoggerFactory;

/***********************************************************************************************************************
 *
 * A partial implementation of {@link ObservationSet} which provides support for listeners and events.
 * 
 * @composed "1" "notify" "0..*" it.tidalwave.observation.event.ObservationListener
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public abstract class ObservationSetSupport implements ObservationSet, Serializable
  {
    private static final Logger log = LoggerFactory.getLogger(ObservationSetSupport.class);
    private static final long serialVersionUID = 546345436345345343L;

    private transient List<ObservationListener> listeners = new ArrayList<ObservationListener>();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static interface ObservationListenerTask
      {
        public void run (@Nonnull ObservationListener listener);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void addListener (final @Nonnull ObservationListener listener)
      {
        if (listeners == null) listeners = new ArrayList<ObservationListener>(); // FIXME: readExternal seems not to work
        listeners.add(listener);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void removeListener (final @Nonnull ObservationListener listener)
      {
        if (listeners == null) listeners = new ArrayList<ObservationListener>(); // FIXME: readExternal seems not to work
        listeners.remove(listener);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void writeExternal (final @Nonnull ObjectOutput oo)
      throws IOException
      {
        // listeners are transient
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void readExternal (final @Nonnull ObjectInput oi)
      throws IOException, ClassNotFoundException
      {
        listeners = new ArrayList<ObservationListener>();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void fireObservationAdded (final @Nonnull Observation observation)
      {
        final ObservationEvent event = new ObservationEvent(this, observation, null);
        fire(new ObservationListenerTask()
          {
            public void run (final @Nonnull ObservationListener listener)
              {
                listener.notifyObservationAdded(event);
              }
          });
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void fireObservationRemoved (final @Nonnull Observation oldObservation)
      {
        final ObservationEvent event = new ObservationEvent(this, null, oldObservation);
        fire(new ObservationListenerTask()
          {
            public void run (final @Nonnull ObservationListener listener)
              {
                listener.notifyObservationRemoved(event);
              }
          });
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void fireObservationChanged (final @Nonnull Observation observation,
                                           final @Nonnull Observation oldObservation)
      {
        final ObservationEvent event = new ObservationEvent(this, observation, oldObservation);
        fire(new ObservationListenerTask()
          {
            public void run (final @Nonnull ObservationListener listener)
              {
                listener.notifyObservationChanged(event);
              }
          });
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void fireNotifyCleared()
      {
        final ObservationEvent event = new ObservationEvent(this, null, null);
        fire(new ObservationListenerTask()
          {
            public void run (final @Nonnull ObservationListener listener)
              {
                listener.notifyCleared(event);
              }
          });
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void fire (final @Nonnull ObservationListenerTask listenerTask)
      {
        for (final ObservationListener listener : new ArrayList<ObservationListener>(listeners))
          {
            try
              {
                listenerTask.run(listener);
              }
            catch (Throwable t)
              {
                log.warn("During event notification: {}", t.toString());
                log.warn("fire()", t);
              }
          }
      }
  }
